"""
This the about file
"""

__all__ = [
    "__title__",
    "__summary__",
    "__uri__",
    "__version__",
    "__author__",
    "__email__",
    "__license__",
    "__copyright__",
]

__title__ = "Harambe"
__version__ = "0.0.9"
__summary__ = "A mid stack Flask based framework that put structure in the " \
              "file system. Features: Caching, Mailing, Storage, CSRF, recaptcha, Cors. " \
              "Supports HTML, Markdown and Jade " \
              "and more"
__uri__ = "https://github.com/mardix/harambe"
__author__ = "Mardix"
__email__ = "mardix@pylot.io"
__license__ = "MIT"
__copyright__ = "(c) 2017 %s" % __author__




import os
import datetime
import uuid
from harambe import (Harambe, models, nav_menu, page_meta, init_app,
                         route, get_config, session, request, redirect,
                         url_for, get, post, flash, flash_data, flash_success, flash_error,
                         abort, recaptcha, storage, get_flash_data)

from flask_login import fresh_login_required
from harambe.exceptions import AppError
from harambe import utils
import json
from . import _del_cache_key
from harambe.contrib.auth import (current_user, authenticated, is_authenticated,
                                not_authenticated, accepts_admin_roles, visible_to_admins)


def main(**kwargs):

    options = kwargs.get("options", {})
    nav_kwargs = kwargs.get("nav_menu", {})

    @accepts_admin_roles
    class AppDataAdmin(Harambe):
        base_route = "/admin"
        decorators = [authenticated, accepts_admin_roles]

        @nav_menu("App Data",
                  visible=visible_to_admins,
                  order=nav_kwargs.pop("order", 100))
        @get("/app-data/")
        def index(self):
            """
            List all posts
            """
            page_meta(title="App Data")
            page = request.args.get("page", 1)

            prefs = models.AppData.query()\
                .order_by(models.AppData.key.asc())\
                .paginate(page=page, per_page=25)

            return {
                "prefs": prefs
            }

        @nav_menu("App Data", visible=False)
        @route("/app-data/<id>/")
        def get(self, id):
            pref = models.AppData.get(id)
            return {
                "pref": pref
            }

        def post(self):
            id = request.form.get("id")
            action = request.form.get("action")
            key = request.form.get("key")
            value = request.form.get("value")
            description = request.form.get("description")

            if action in ["update", "delete"]:
                pref = models.AppData.get(id)
                if not pref:
                    abort(404, "App Data doesn't exist")

                if action == "update":
                    if pref.is_editable:
                        pref.data = value
                        ns, k = pref.key.split(".", 2)
                        _del_cache_key(namespace=ns, key=k)

                    pref.update(description=description)
                    flash_success("App Data '%s' updated successfully!" % pref.key)
                elif action == "delete":
                    flash_success("App Data '%s' deleted successfully!" % pref.key)
                    return redirect(self.index)

            return redirect(self.get, id=id)

    return AppDataAdmin



import logging
import sys
import inspect
from harambe import (db,
                   utils,
                   cache,
                   models)
from sqlalchemy.exc import SQLAlchemyError


def make_key(key):
    return utils.slugify(key)


class ApplicationOption(db.Model):
    """
    Application Option Model
    Allow to save app specific data
    """
    description = db.Column(db.String(250))
    key = db.Column(db.String(250), index=True, unique=True)
    value = db.Column(db.JSONType())
    autoload = db.Column(db.Boolean, default=False)


    @classmethod
    def get_by_key(cls, key):
        key = make_key(key)
        return cls.query().filter(key == key).first()

    @classmethod
    def get_key(cls, key):
        d = cls.get_by_key(key)
        return {} if not d else utils.dict_dot(d.value)

    @classmethod
    def set_key(cls, key, value, description=None, autoload=False):
        if not isinstance(value, dict):
            raise ValueError("AppSettings value must be a dict")
        appset = cls.get_by_key(key)
        if appset:
            data = appset.value
            data.update(value)
            upd = {
                "value": data
            }
            if description:
                upd["description"] = description
            appset.update(upd)
        else:
            cls.create(key=make_key(key),
                       value=value,
                       description=description,
                       autoload=autoload)


class AppOption(object):
    """
    Utility to access app option
    """
    def __init__(self, key, cache=True, cache_ttl=300):

        """
        :param key: string or __name__ (of the current module)
        :param cache: bool - If true it will cache the data
        :param cache_ttl: Time for the cache to live
        """
        self.key = make_key(key)
        self.cache = cache
        self.cache_ttl = cache_ttl
        self.cache_key = "__SHAFTAPPOPTION__:%s" % self.key

    def init(self, value, description, autoload=False):
        """
        Initialize a new entry if it doesn't exist
        :param value: dict
        :param description: string
        :param autoload: bool
        :return:
        """
        try:
            if not models.ApplicationOption.get_by_key(self.key):
                models.ApplicationOption.set_key(key=self.key,
                                                 value=value,
                                                 description=description,
                                                 autoload=autoload)
        except SQLAlchemyError as e:
            return None

    def set(self, value):
        """
        Set/Update data
        :param value: dict
        :return:
        """
        try:
            self._del_cache()
            models.ApplicationOption.set_key(key=self.key, value=value)
        except SQLAlchemyError as e:
            return None

    def get(self, name=None, default=None):
        """
        Get the data
        :param name: string or dot notation
        :param default: the default value if None
        :return: mixed
        """
        if not self.cache:
            return self._get()

        res = cache.get(self.cache_key)
        if not res:
            res = self._get()
            cache.set(self.cache_key, res, self.cache_ttl)
        return res.get(name, default) if name else res

    def _get(self):
        """
        Get the key from DB
        :return:
        """
        try:
            return models.ApplicationOption.get_key(key=self.key)
        except SQLAlchemyError as e:
            return None

    def _del_cache(self):
        """
        Delete the cache
        :return:
        """
        cache.delete(self.cache_key)


"""
A jade extension that embed jade markup in the .html file

{% jade %}


{% endjade %}

"""

import re
import pyjade
from jinja2.ext import Extension
from pyjade.utils import process
from pyjade.ext.jinja import Compiler
from jinja2 import (TemplateSyntaxError, nodes)


begin_tag_rx = r'\{%\-?\s*jade.*?%\}'
end_tag_rx = r'\{%\-?\s*endjade\s*\-?%\}'

begin_tag_m = re.compile(begin_tag_rx)
end_tag_m = re.compile(end_tag_rx)


def convert(text, filename=None):
    return process(text, filename=filename, compiler=Compiler)

class TemplateIndentationError(TemplateSyntaxError): pass


class JadeTagExtension(Extension):
    tags = set(['jade'])

    def _get_lineno(self, source):
        matches = re.finditer(r"\n", source)
        if matches:
            return len(tuple(matches))
        return 0

    def preprocess(self, source, name, filename=None):
        ret_source = ''
        start_pos = 0

        while True:
            tag_match = begin_tag_m.search(source, start_pos)
            if tag_match:
                end_tag = end_tag_m.search(source, tag_match.end())
                if not end_tag:
                    raise TemplateSyntaxError('Expecting "endjade" tag',
                                              self._get_lineno(source[:start_pos]))

                jade_source = source[tag_match.end(): end_tag.start()]
                jade_source = convert(jade_source)

                try:
                    ret_source += source[start_pos: tag_match.start()] + jade_source
                except TemplateIndentationError as e:
                    raise TemplateSyntaxError(e.message, e.lineno, name=name, filename=filename)
                except TemplateSyntaxError as e:
                    raise TemplateSyntaxError(e.message, e.lineno, name=name, filename=filename)
                start_pos = end_tag.end()
            else:
                ret_source += source[start_pos:]
                break

        return ret_source





"""
Harambe

"""

import re
import os
import sys
import arrow
import jinja2
import inspect
import logging
import functools
import pkg_resources
import logging.config
from six import string_types
from werkzeug import import_string
from flask_assets import Environment
from werkzeug.contrib.fixers import ProxyFix
from werkzeug.routing import (BaseConverter,
                              parse_rule)
from flask import (Flask,
                   g,
                   render_template,
                   flash,
                   session,
                   make_response,
                   Response,
                   request,
                   abort,
                   url_for as f_url_for,
                   redirect as f_redirect)
from . import (utils, exceptions)
from .extras.harambealch import HarambeAlchemy
from .__about__ import *

_py2 = sys.version_info[0] == 2

# ------------------------------------------------------------------------------

__all__ = [
            "Harambe",
            "GoHarambe",
            "db",
            "models",
            "views",
            "get_env",
            "set_env",
            "get_app_env",
            "get_env_config",
            "get_config",
            "page_meta",
            "flash_success",
            "flash_error",
            "flash_info",
            "flash_data",
            "get_flash_data",
            "init_app",
            "register_package",
            "register_models",
            "utc_now",
            "local_now",
            "to_local_datetime",

            # For convenience when importing from harambe, but can use
            # the flask one
            "flash",
            "session",
            "request",
            "abort",
            "g",

            # They have been altered with extra functionalities
            "redirect",
            "url_for"
           ]


# Hold the current environment
__ENV__ = None

is_method = lambda x: inspect.ismethod if _py2 else inspect.isfunction

# Will hold all active class views
# It can be used for redirection etc
# ie: redirect(views.ContactPage.index)
views = type('', (), {})

# Will hold models from apps, or to be shared
# ie, set new model property -> models.MyNewModel = MyModel
# ie: use property -> models.MyNewModel.all()
# For convenience, use `register_models(**kw)` to register the models
# By default harambe will load all the application/models.py models
models = type('', (), {})


# Setup the DB
# upon initialization will use the right URI for it
# also, it exposes the db object to all modules
db = HarambeAlchemy()


def register_models(**kwargs):
    """
    Alias to register model
    :param kwargs:
    :return:
    """
    [setattr(models, k, v) for k, v in kwargs.items()]


def set_env(env):
    """
    Set the envrionment manually
    :param env:
    :return:
    """
    global __ENV__
    __ENV__ = env.lower().capitalize()


def get_env():
    """
    Return the Capitalize environment name
    It can be used to retrieve class base config
    Default: Development
    :returns: str Capitalized
    """
    if not __ENV__:
        env = os.environ["env"] if "env" in os.environ else "Dev"
        set_env(env)
    return __ENV__


def get_app_env():
    """
    if the app and the envi are passed in the command line as 'app=$app:$env'
    :return: tuple app, env
    """
    app, env = None, get_env()
    if "app" in os.environ:
        app = os.environ["app"].lower()
        if ":" in app:
            app, env = os.environ["app"].split(":", 2)
            set_env(env)
    return app, env


def get_env_config(config):
    """
    Return config class based based on the config
    :param config : Object - The configuration module containing the environment object
    """
    return getattr(config, get_env())


def init_app(kls):
    """
    To bind middlewares, plugins that needs the 'app' object to init
    Bound middlewares will be assigned on cls.init()
    """
    if not hasattr(kls, "__call__"):
        raise exceptions.HarambeError("init_app: '%s' is not callable" % kls)
    Harambe._init_apps.add(kls)
    return kls


def register_package(pkg):
    """
    Allow to register an app packages by loading and exposing: templates, static,
    and exceptions for abort()

    Structure of package
        root
            | $package_name
                | __init__.py
                |
                | /templates
                    |
                    |
                |
                | /static
                    |
                    | assets.yml

    :param pkg: str - __package__
                    or __name__
                    or The root dir
                    or the dotted resource package (package.path.path,
                    usually __name__ of templates and static
    """

    root_pkg_dir = pkg
    if not os.path.isdir(pkg) and "." in pkg:
        root_pkg_dir = pkg_resources.resource_filename(pkg, "")

    template_path = os.path.join(root_pkg_dir, "templates")
    static_path = os.path.join(root_pkg_dir, "static")

    logging.info("Registering App: " + pkg)
    if os.path.isdir(template_path):
        template_path = jinja2.FileSystemLoader(template_path)
        Harambe._template_paths.add(template_path)

    if os.path.isdir(static_path):
        Harambe._static_paths.add(static_path)
        Harambe._add_asset_bundle(static_path)


def get_config(key, default=None):
    """
    Shortcut to access the application's config in your class
    :param key: The key to access
    :param default: The default value when None
    :returns mixed:
    """
    return Harambe._app.config.get(key, default) if Harambe._app else default


def page_meta(title=None, **kwargs):
    """
    Meta allows you to add page meta data in the request `g` context
    :params **kwargs:

    meta keys we're expecting:
        title (str)
        description (str)
        url (str) (Will pick it up by itself if not set)
        image (str)
        site_name (str) (but can pick it up from config file)
        object_type (str)
        keywords (list)
        locale (str)
        card (str)

        **Boolean By default these keys are True
        use_opengraph
        use_twitter
        use_googleplus
python
    """
    default = dict(
            title="",
            description="",
            url="",
            image="",
            site_name="",
            object_type="article",
            locale="",
            keywords=[],
            use_opengraph=True,
            use_googleplus=True,
            use_twitter=True,
            properties={}
        )
    meta = getattr(g, "__META__", default)
    if title:
        kwargs["title"] = title
    meta.update(**kwargs)
    setattr(g, "__META__", meta)


def flash_success(msg):
    """
    Alias to flash, but set a success message
    :param msg:
    :return:
    """
    return flash(msg, "success")


def flash_error(msg):
    """
    Alias to flash, but set an error message
    :param msg:
    :return:
    """
    return flash(msg, "error")


def flash_info(msg):
    """
    Alias to flash, but set an info message
    :param msg:
    :return:
    """
    return flash(msg, "info")


def flash_data(data):
    """
    Just like flash, but will save data
    :param data:
    :return:
    """
    session["_flash_data"] = data


def get_flash_data():
    """
    Retrieved
    :return: mixed
    """
    return session.pop("_flash_data", None)


def utc_now():
    """
    Return the utcnow arrow object
    :return: Arrow
    """
    return arrow.utcnow()


def local_now():
    return to_local_datetime(utc_now())


def to_local_datetime(utcdatetime, timezone=None):
    """
    Return local datetime based on the timezone
    :param utcdatetime: Arrow or string
    :param timezone:
    :return:
    """
    timezone = timezone or get_config("DATETIME_LOCAL_TIMEZONE", "US/Eastern")
    if isinstance(utcdatetime, arrow.Arrow):
        return utcdatetime.to(timezone)
    return arrow.get(utcdatetime, timezone)


# ------------------------------------------------------------------------------
# Altered flask functions


def url_for(endpoint, **kw):
    """
    Harambe url_for is an alias to the flask url_for, with the ability of
    passing the function signature to build the url, without knowing the endpoint
    :param endpoint:
    :param kw:
    :return:
    """

    _endpoint = None
    if isinstance(endpoint, string_types):
        return f_url_for(endpoint, **kw)
    else:
        # self, will refer the caller method, by getting the method name
        if isinstance(endpoint, Harambe):
            fn = sys._getframe().f_back.f_code.co_name
            endpoint = getattr(endpoint, fn)

        if is_method(endpoint):
            _endpoint = _get_action_endpoint(endpoint)
            if not _endpoint:
                _endpoint = _build_endpoint_route_name(endpoint)
    if _endpoint:
        return f_url_for(_endpoint, **kw)
    else:
        raise exceptions.HarambeError('Harambe `url_for` received an invalid endpoint')


def redirect(endpoint, **kw):
    """
    Redirect allow to redirect dynamically using the classes methods without
    knowing the right endpoint.
    Expecting all endpoint have GET as method, it will try to pick the first
    match, based on the endpoint provided or the based on the Rule map_url

    An endpoint can also be passed along with **kw

    An http: or https: can also be passed, and will redirect to that site.

    example:
        redirect(self.hello_world)
        redirect(self.other_page, name="x", value="v")
        redirect("https://google.com")
        redirect(views.ContactPage.index)
    :param endpoint:
    :return: redirect url
    """

    _endpoint = None

    if isinstance(endpoint, string_types):
        _endpoint = endpoint
        # valid for https:// or /path/
        # Endpoint should not have slashes. Use : (colon) to build endpoint
        if "/" in endpoint:
            return f_redirect(endpoint)
        else:
            for r in Harambe._app.url_map.iter_rules():
                _endpoint = endpoint
                if 'GET' in r.methods and endpoint in r.endpoint:
                    _endpoint = r.endpoint
                    break
    else:
        # self, will refer the caller method, by getting the method name
        if isinstance(endpoint, Harambe):
            fn = sys._getframe().f_back.f_code.co_name
            endpoint = getattr(endpoint, fn)

        if is_method(endpoint):
            _endpoint = _get_action_endpoint(endpoint)
            if not _endpoint:
                _endpoint = _build_endpoint_route_name(endpoint)
    if _endpoint:
        return f_redirect(url_for(_endpoint, **kw))
    else:
        raise exceptions.HarambeError("Invalid endpoint")


def _get_action_endpoint(action):
    """
    Return the endpoint base on the view's action
    :param action:
    :return:
    """
    _endpoint = None
    if is_method(action):
        if hasattr(action, "_rule_cache"):
            rc = action._rule_cache
            if rc:
                k = list(rc.keys())[0]
                rules = rc[k]
                len_rules = len(rules)
                if len_rules == 1:
                    rc_kw = rules[0][1]
                    _endpoint = rc_kw.get("endpoint", None)
                    if not _endpoint:
                        _endpoint = _build_endpoint_route_name(action)
                elif len_rules > 1:
                    _prefix = _build_endpoint_route_name(action)
                    for r in Harambe._app.url_map.iter_rules():
                        if ('GET' in r.methods or 'POST' in r.methods) \
                                and _prefix in r.endpoint:
                            _endpoint = r.endpoint
                            break
    return _endpoint


def _build_endpoint_route_name(endpoint):

    is_class = inspect.isclass(endpoint)
    class_name = endpoint.im_class.__name__ if not is_class else endpoint.__name__
    method_name = endpoint.__name__

    cls = endpoint.im_class() \
        if (not hasattr(endpoint, "__self__") or endpoint.__self__ is None) \
        else endpoint.__self__

    return build_endpoint_route_name(cls, method_name, class_name)

# ------------------------------------------------------------------------------


class Harambe(object):
    decorators = []
    base_route = None
    route_prefix = None
    trailing_slash = True
    base_layout = "layouts/base.jade"
    template_markup = "jade"
    assets = None
    logger = None
    _ext = set()
    __special_methods = ["get", "put", "patch", "post", "delete", "index"]
    _installed_apps = []
    _app = None
    _init_apps = set()
    _template_paths = set()
    _static_paths = set()
    _asset_bundles = set()

    @classmethod
    def __call__(cls,
                 flask_or_import_name,
                 projects=None,
                 project_name=None,
                 app_directory=None,
                 ):
        """

        :param flask_or_import_name: Flask instance or import name -> __name__
        :param projects: dict of app and views to load. ie:
            {
                "main": [
                    "views.main",
                    "views.api"
                ]
            }
        :param project_name: name of the project. If empty, it will try to get
                             it from the app_env(). By default it is "main"
                             The app main is set as environment variable
                             ie: app=PROJECT_NAME:CONFIG -> app=main:production
        :param app_directory: the directory name relative to the current execution path
        :return:
        """
        if not app_directory:
            app_directory = "app"
        if not project_name:
            project_name = get_app_env()[0] or "main"

        app_env = get_env()

        app = flask_or_import_name \
            if isinstance(flask_or_import_name, Flask) \
            else Flask(flask_or_import_name)

        app.wsgi_app = ProxyFix(app.wsgi_app)
        app.url_map.converters['regex'] = RegexConverter
        app.template_folder = "%s/templates" % app_directory
        app.static_folder = "%s/static" % app_directory

        # Load configs
        c = "%s.config.%s" % (app_directory, app_env)
        app.config.from_object(c)

        cls._app = app
        cls.assets = Environment(cls._app)
        cls._load_extensions()
        cls._setup_logger()
        cls._setup_db()

        cls.setup_installed_apps()
        cls._expose_models()

        try:
            # import models
            m = "%s.models" % app_directory
            import_string(m)
            cls._expose_models()

            # import projects views
            if not projects:
                projects = {"main": "main"}

            if project_name not in projects:
                raise ValueError("Missing project: %s" % project_name)

            _projects = projects.get(project_name)
            if isinstance(_projects, string_types):
                _projects = [_projects]
            for _ in _projects:
                import_string("%s.views.%s" % (app_directory, _))

        except ImportError as ie1:
            pass
        cls._expose_models()

        # Setup init_app
        # init_app instanciate functions that may need the flask.app object
        # Usually for flask extension to be setup
        _ = [_app(cls._app) for _app in cls._init_apps]

        # Add bundles
        cls._add_asset_bundle(cls._app.static_folder)

        # Register templates
        if cls._template_paths:
            loader = [cls._app.jinja_loader] + list(cls._template_paths)
            cls._app.jinja_loader = jinja2.ChoiceLoader(loader)

        # Static
        if cls._static_paths:
            cls.assets.load_path = [cls._app.static_folder] + list(cls._static_paths)
            [cls.assets.from_yaml(a) for a in cls._asset_bundles]

        # Register views
        for subcls in cls.__subclasses__():
            base_route = subcls.base_route
            if not base_route:
                base_route = utils.dasherize(utils.underscore(subcls.__name__))
                if subcls.__name__.lower() == "index":
                    base_route = "/"
            subcls._register(cls._app, base_route=base_route)

        @cls._app.after_request
        def _after_request_cleanup(response):
            #cls._global["__META__"] = cls._default_page_meta.copy()
            return response

        return cls._app

    @classmethod
    def setup_installed_apps(cls):
        """
        To import 3rd party applications along with associated properties

        It is a list of dict or string.

        When a dict, it contains the `app` key and the configuration,
        if it's a string, it is just the app name

        If you require dependencies from other packages, dependencies
        must be placed before the calling package.

        It is required that __init__ in the package app has an entry point method
        -> 'main(**kw)' which will be used to setup the default app.

        As a dict
        INSTALLED_APPS = [
            "it.can.be.a.string.to.the.module",
            ("in.a.tuple.with.props.dict", {options}),
            [
                ("multi.app.list.in.a.list.of.tuple", {options}),
                ("multi.app.list.in.a.list.of.tuple2", {options})
            ]
        ]

        :return:
        """

        cls._installed_apps = cls._app.config.get("INSTALLED_APPS", [])
        if cls._installed_apps:
            def import_app(module, props={}):
                _ = import_string(module)
                setattr(_, "__options__", utils.dict_dot(props))

            for k in cls._installed_apps:
                if isinstance(k, string_types): # One string
                    import_app(k, {})
                elif isinstance(k, tuple):
                    import_app(k[0], k[1])
                elif isinstance(k, list):  # list of tuple[(module props), ...]
                    for t in k:
                        import_app(t[0], t[1])


    @classmethod
    def render(cls, data={}, _template=None, _layout=None, **kwargs):
        """
        Render the view template based on the class and the method being invoked
        :param data: The context data to pass to the template
        :param _template: The file template to use. By default it will map the module/classname/action.html
        :param _layout: The body layout, must contain {% include __template__ %}
        """

        # Invoke the page meta so it can always be set
        page_meta()

        # Add some global Harambe data in g, along with APPLICATION DATA
        vars = dict(
            __NAME__=__title__,
            __VERSION__=__version__,
            __YEAR__=utc_now().year
        )
        for k, v in vars.items():
            setattr(g, k, v)

        if not _template:
            stack = inspect.stack()[1]
            action_name = stack[3]
            _template = build_endpoint_route_name(cls, action_name)
            _template = utils.list_replace([".", ":"], "/", _template)
            _template = "%s.%s" % (_template, cls.template_markup)

        data = data or dict()
        data.update(kwargs)
        data["__template__"] = _template

        return render_template(_layout or cls.base_layout, **data)

    @classmethod
    def _add_asset_bundle(cls, path):
        """
        Add a webassets bundle yml file
        """
        f = "%s/assets.yml" % path
        if os.path.isfile(f):
            cls._asset_bundles.add(f)

    @classmethod
    def _setup_logger(cls):
        logging_config = cls._app.config.get("LOGGING")
        if not logging_config:
            logging_config = {
                "version": 1,
                "handlers": {
                    "default": {
                        "class": cls._app.config.get("LOGGING_CLASS", "logging.StreamHandler")
                    }
                },
                'loggers': {
                    '': {
                        'handlers': ['default'],
                        'level': 'WARN',
                    }
                }
            }

        logging.config.dictConfig(logging_config)
        cls.logger = logging.getLogger("root")
        cls._app._logger = cls.logger
        cls._app._loger_name = cls.logger.name

    @classmethod
    def _setup_db(cls):
        cls._app.db = None
        uri = cls._app.config.get("DB_URI")
        if uri:
            db._connect(uri, cls._app)
            cls._app.db = db

    @classmethod
    def _expose_models(cls):
        if cls._app.db:
            register_models(**{m.__name__:m
                               for m in cls._app.db.Model.__subclasses__()
                               if not hasattr(models, m.__name__)})

    @classmethod
    def _register(cls,
                  app,
                  base_route=None,
                  subdomain=None,
                  route_prefix=None,
                  trailing_slash=None):
        """Registers a Harambe class for use with a specific instance of a
        Flask app. Any methods not prefixes with an underscore are candidates
        to be routed and will have routes registered when this method is
        called.

        :param app: an instance of a Flask application

        :param base_route: The base path to use for all routes registered for
                           this class. Overrides the base_route attribute if
                           it has been set.

        :param subdomain:  A subdomain that this registration should use when
                           configuring routes.

        :param route_prefix: A prefix to be applied to all routes registered
                             for this class. Precedes base_route. Overrides
                             the class' route_prefix if it has been set.
        """

        if cls is Harambe:
            raise TypeError("cls must be a subclass of Harambe, not Harambe itself")

        # Create a unique namespaced key to access view.
        # $module.$class_name.$Method
        module = cls.__module__.split(".")[-1]

        if not hasattr(views, module):
            setattr(views, module, type('', (), {}))
        mod = getattr(views, module)
        setattr(mod, cls.__name__, cls)

        if base_route:
            cls.orig_base_route = cls.base_route
            cls.base_route = base_route

        if route_prefix:
            cls.orig_route_prefix = cls.route_prefix
            cls.route_prefix = route_prefix

        if not subdomain:
            if hasattr(app, "subdomain") and app.subdomain is not None:
                subdomain = app.subdomain
            elif hasattr(cls, "subdomain"):
                subdomain = cls.subdomain

        if trailing_slash is not None:
            cls.orig_trailing_slash = cls.trailing_slash
            cls.trailing_slash = trailing_slash

        for name, value in get_interesting_members(Harambe, cls):
            proxy = cls.make_proxy_method(name)
            route_name = build_endpoint_route_name(cls, name)
            try:
                if hasattr(value, "_rule_cache") and name in value._rule_cache:
                    for idx, cached_rule in enumerate(value._rule_cache[name]):
                        rule, options = cached_rule
                        rule = cls.build_rule(rule)
                        sub, ep, options = cls.parse_options(options)

                        if not subdomain and sub:
                            subdomain = sub

                        if ep:
                            endpoint = ep
                        elif len(value._rule_cache[name]) == 1:
                            endpoint = route_name
                        else:
                            endpoint = "%s_%d" % (route_name, idx,)

                        app.add_url_rule(rule, endpoint, proxy,
                                         subdomain=subdomain,
                                         **options)
                elif name in cls.__special_methods:
                    if name in ["get", "index"]:
                        methods = ["GET"]
                        if name == "index":
                            if hasattr(value, "_methods_cache"):
                                methods = value._methods_cache
                    else:
                        methods = [name.upper()]

                    rule = cls.build_rule("/", value)
                    if not cls.trailing_slash:
                        rule = rule.rstrip("/")
                    app.add_url_rule(rule, route_name, proxy,
                                     methods=methods,
                                     subdomain=subdomain)

                else:
                    methods = value._methods_cache \
                        if hasattr(value, "_methods_cache") \
                        else ["GET"]

                    name = utils.dasherize(name)
                    route_str = '/%s/' % name
                    if not cls.trailing_slash:
                        route_str = route_str.rstrip('/')
                    rule = cls.build_rule(route_str, value)
                    app.add_url_rule(rule, route_name, proxy,
                                     subdomain=subdomain,
                                     methods=methods)
            except DecoratorCompatibilityError:
                raise DecoratorCompatibilityError("Incompatible decorator detected on %s in class %s" % (name, cls.__name__))

        if hasattr(cls, "orig_base_route"):
            cls.base_route = cls.orig_base_route
            del cls.orig_base_route

        if hasattr(cls, "orig_route_prefix"):
            cls.route_prefix = cls.orig_route_prefix
            del cls.orig_route_prefix

        if hasattr(cls, "orig_trailing_slash"):
            cls.trailing_slash = cls.orig_trailing_slash
            del cls.orig_trailing_slash

    @classmethod
    def parse_options(cls, options):
        """Extracts subdomain and endpoint values from the options dict and returns
           them along with a new dict without those values.
        """
        options = options.copy()
        subdomain = options.pop('subdomain', None)
        endpoint = options.pop('endpoint', None)
        return subdomain, endpoint, options,

    @classmethod
    def make_proxy_method(cls, name):
        """Creates a proxy function that can be used by Flasks routing. The
        proxy instantiates the Harambe subclass and calls the appropriate
        method.
        :param name: the name of the method to create a proxy for
        """

        i = cls()
        view = getattr(i, name)

        for decorator in cls.decorators:
            view = decorator(view)

        @functools.wraps(view)
        def proxy(**forgettable_view_args):
            # Always use the global request object's view_args, because they
            # can be modified by intervening function before an endpoint or
            # wrapper gets called. This matches Flask's behavior.
            del forgettable_view_args

            if hasattr(i, "before_request"):
                response = i.before_request(name, **request.view_args)
                if response is not None:
                    return response

            before_view_name = "before_" + name
            if hasattr(i, before_view_name):
                before_view = getattr(i, before_view_name)
                response = before_view(**request.view_args)
                if response is not None:
                    return response

            response = view(**request.view_args)

            # You can also return a dict or None, it will pass it to render
            if isinstance(response, dict) or response is None:
                response = response or {}
                if hasattr(i, "_renderer"):
                    response = i._renderer(response)
                else:
                    _template = build_endpoint_route_name(cls, view.__name__)
                    _template = utils.list_replace([".", ":"], "/", _template)
                    _template = "%s.%s" % (_template, cls.template_markup)
                    response.setdefault("_template", _template)
                    response = i.render(**response)

            if not isinstance(response, Response):
                response = make_response(response)

            for ext in cls._ext:
                response = ext(response)

            after_view_name = "after_" + name
            if hasattr(i, after_view_name):
                after_view = getattr(i, after_view_name)
                response = after_view(response)

            if hasattr(i, "after_request"):
                response = i.after_request(name, response)

            return response

        return proxy

    @classmethod
    def build_rule(cls, rule, method=None):
        """Creates a routing rule based on either the class name (minus the
        'View' suffix) or the defined `base_route` attribute of the class

        :param rule: the path portion that should be appended to the
                     route base

        :param method: if a method's arguments should be considered when
                       constructing the rule, provide a reference to the
                       method here. arguments named "self" will be ignored
        """

        rule_parts = []

        if cls.route_prefix:
            rule_parts.append(cls.route_prefix)

        base_route = cls.get_base_route()
        if base_route:
            rule_parts.append(base_route)

        rule_parts.append(rule)
        ignored_rule_args = ['self']
        if hasattr(cls, 'base_args'):
            ignored_rule_args += cls.base_args

        if method:
            args = get_true_argspec(method)[0]
            for arg in args:
                if arg not in ignored_rule_args:
                    rule_parts.append("<%s>" % arg)

        result = "/%s" % "/".join(rule_parts)
        return re.sub(r'(/)\1+', r'\1', result)

    @classmethod
    def get_base_route(cls):
        """Returns the route base to use for the current class."""
        base_route = cls.__name__.lower()
        if cls.base_route is not None:
            base_route = cls.base_route
            base_rule = parse_rule(base_route)
            cls.base_args = [r[2] for r in base_rule]
        return base_route.strip("/")

    @staticmethod
    def _bind_route_rule_cache(f, rule, append_method=False, **kwargs):
        # Put the rule cache on the method itself instead of globally
        if rule is None:
            rule = utils.dasherize(f.__name__) + "/"
        if not hasattr(f, '_rule_cache') or f._rule_cache is None:
            f._rule_cache = {f.__name__: [(rule, kwargs)]}
        elif not f.__name__ in f._rule_cache:
            f._rule_cache[f.__name__] = [(rule, kwargs)]
        else:
            # when and endpoint accepts multiple METHODS, ie: post(), get()
            if append_method:
                for r in f._rule_cache[f.__name__]:
                    if r[0] == rule and "methods" in r[1] and "methods" in kwargs:
                        r[1]["methods"] = list(set(r[1]["methods"] + kwargs["methods"]))
            else:
                f._rule_cache[f.__name__].append((rule, kwargs))
        return f

    @classmethod
    def _load_extensions(cls):
        extensions = [
            'pyjade.ext.jinja.PyJadeExtension',
            'harambe.extras.jade.JadeTagExtension',
            'harambe.extras.md.MarkdownExtension',
            'harambe.extras.md.MarkdownTagExtension',
        ]
        if cls._app.config.get("COMPRESS_HTML"):
            extensions.append('harambe.extras.htmlcompress.HTMLCompress')
        for ext in extensions:
            cls._app.jinja_env.add_extension(ext)

# GoHarambe
GoHarambe = Harambe()

# ------------------------------------------------------------------------------

def build_endpoint_route_name(cls, method_name, class_name=None):
    """
    Build the route endpoint
    It is recommended to place your views in /views directory, so it can build
    the endpoint from it. If not, it will make the endpoint from the module name
    The main reason for having the views directory, it is explicitely easy
    to see the path of the view

    :param cls: The view class
    :param method_name: The name of the method
    :param class_name: To pass the class name.
    :return: string
    """

    module = cls.__module__.split("views.")[1] if ".views." in cls.__module__ \
        else cls.__module__.split(".")[-1]
    return "%s.%s:%s" % (module, class_name or cls.__name__, method_name)


def get_interesting_members(base_class, cls):
    """Returns a generator of methods that can be routed to"""

    base_members = dir(base_class)
    predicate = inspect.ismethod if _py2 else inspect.isfunction
    all_members = inspect.getmembers(cls, predicate=predicate)
    return (member for member in all_members
            if not member[0] in base_members
            and ((hasattr(member[1], "__self__") and not member[1].__self__ in inspect.getmro(cls)) if _py2 else True)
            and not member[0].startswith("_")
            and not member[0].startswith("before_")
            and not member[0].startswith("after_"))


def apply_function_to_members(cls, fn):
    for name, method in get_interesting_members(Harambe, cls):
        setattr(cls, name, fn(method))


def get_true_argspec(method):
    """Drills through layers of decorators attempting to locate the actual argspec for the method."""

    argspec = inspect.getargspec(method)
    args = argspec[0]
    if args and args[0] == 'self':
        return argspec
    if hasattr(method, '__func__'):
        method = method.__func__
    if not hasattr(method, '__closure__') or method.__closure__ is None:
        raise DecoratorCompatibilityError

    closure = method.__closure__
    for cell in closure:
        inner_method = cell.cell_contents
        if inner_method is method:
            continue
        if not inspect.isfunction(inner_method) \
            and not inspect.ismethod(inner_method):
            continue
        true_argspec = get_true_argspec(inner_method)
        if true_argspec:
            return true_argspec


class DecoratorCompatibilityError(Exception):
    pass


class RegexConverter(BaseConverter):
    def __init__(self, url_map, *items):
        super(RegexConverter, self).__init__(url_map)
        self.regex = items[0]






"""
Harambe Auth
"""

import logging
from . import signals, exceptions, oauth
import flask_login
from flask import current_app
from harambe.exceptions import AppError
from harambe import (_,
                   utc_now,
                   get_config,
                   abort,
                   send_mail,
                   url_for,
                   views,
                   models,
                   utils,
                   request,
                   redirect,
                   flash,
                   session,
                   init_app)
import models as auth_models
from flask_login import current_user

__version__ = "1.0.0"

__options__ = utils.dict_dot({})


# USER ROLES
ROLES_SUPERADMIN = ["SUPERADMIN"]
ROLES_ADMIN = ROLES_SUPERADMIN + ["ADMIN"]
ROLES_MANAGER = ROLES_ADMIN + ["MANAGER"]
ROLES_CONTRIBUTOR = ROLES_MANAGER + ["EDITOR", "CONTRIBUTOR"]
ROLES_MODERATOR = ROLES_CONTRIBUTOR + ["MODERATOR"]

# LOGIN MANAGER: Flask Login
login_manager = flask_login.LoginManager()
login_manager.login_message_category = "error"
init_app(login_manager.init_app)


@login_manager.user_loader
def load_user(userid):
    return get_user_by_id(userid)


def ___main(app):
    @app.before_request
    def force_password_change():
        _ = __options__.get("require_password_change_exclude_endpoints")
        _ = [] if not isinstance(_, list) else _

        exclude_endpoints = ["static", "ContactPage:index", "Index:index",
                             "AuthLogin:logout"] + _

        if current_user and current_user.is_authenticated:
            if request.endpoint \
                    and request.endpoint not in exclude_endpoints:
                if request.endpoint != "AuthAccount:change_password" \
                        and session_get_require_password_change():
                    flash("Password Change is required", "info")
                    return redirect(views.AuthAccount.change_password)


# ------------------------------------------------------------------------------

def is_authenticated():
    """ A shortcut to check if a user is authenticated """
    return current_user and current_user.is_authenticated and current_user.is_active


def not_authenticated():
    """ A shortcut to check if user not authenticated."""
    return not is_authenticated()


def get_random_password(length=8):
    return utils.generate_random_string(length)


# ------------------------------------------------------------------------------
# VISIBILITY:
# The methods below return bool that are meant be pass in the
# nav_title(visible=fn) `visible` args
#

def visible_to_roles(*roles):
    """
    This is a @nav_title specific function to set the visibility of menu based on
    roles
    :param roles:
    :return: callback fn
    """
    if is_authenticated():
        return True if current_user.has_any_roles(*roles) else False
    return False


# Alias
def visible_to_superadmins():
    return visible_to_roles(*ROLES_SUPERADMIN)


def visible_to_admins():
    return visible_to_roles(*ROLES_ADMIN)


def visible_to_managers():
    return visible_to_roles(*ROLES_MANAGER)


def visible_to_contributors():
    return visible_to_roles(*ROLES_CONTRIBUTOR)


def visible_to_moderators():
    return visible_to_roles(*ROLES_MODERATOR)


def visible_to_authenticated():
    return is_authenticated()


def visible_to_non_authenticated():
    return not_authenticated()

# ------------------------------------------------------------------------------

# TOKENIZATION


def get_jwt_secret():
    """
    Get the JWT secret
    :return: str
    """
    secret_key = __options__.get("jwt_secret") or get_config("JWT_SECRET") or get_config("SECRET_KEY")
    if not secret_key:
        raise exceptions.AuthError("Missing config JWT/SECRET_KEY")
    return secret_key


def get_jwt_salt():
    """
    Get the JWT salt
    :return: str
    """
    return __options__.get("jwt_salt", "harambe:contrib:aut")


def get_jwt_ttl():
    """
    Get JWT time to live
    :return:
    """
    return __options__.get("jwt_ttl", 3600)


# ------------------------------------------------------------------------------
# SIGNUP + LOGIN


def _user(user):
    """
    Factory function to AuthUser
    :param user: AuthUser
    :return:
    """
    return UserModel(user) if user else None


def create_user(username, password, email=None, first_name=None, last_name=None,
                role="MEMBER", login_method=None):
    """
    Create a new user
    :param username:
    :param password:
    :param email:
    :param first_name:
    :param last_name:
    :param role: str
    :return: AuthUser
    """

    def cb():
        return _user(models.AuthUser.new(username=username,
                                         password=password,
                                         email=email,
                                         first_name=first_name,
                                         last_name=last_name,
                                         login_method=login_method,
                                         role=role))
    return signals.on_create_user(cb)


def get_user_by_id(id):
    """
    To get a user by id
    :param id: int
    :return: AuthUser
    """
    return _user(models.AuthUser.get(id))


def get_user_by_username(username):
    """
    Return AuthUser by username
    :param username: 
    :return: AuthUser
    """
    return _user(models.AuthUser.get_by_username(username))


def get_user_by_email(email):
    """
    Return AuthUser by email
    :param email:
    :return: AuthUser
    """
    return _user(models.AuthUser.get_by_email(email))


def get_user_by_jwt(token=None):
    """
    Return the AuthUser associated to the token, otherwise it will return None.
    If token is not provided, it will pull it from the headers: Authorization

    Exception:
    Along with AuthError, it may
    :param token:
    :return: AuthUser
    """

    if not token:
        if not 'Authorization' in request.headers:
            raise exceptions.AuthError("Missing Authorization Bearer in headers")
        data = request.headers['Authorization'].encode('ascii', 'ignore')
        token = str.replace(str(data), 'Bearer ', '').strip()

    secret_key = get_jwt_secret()

    s = utils.unsign_jwt(token=token,
                         secret_key=secret_key,
                         salt=get_jwt_salt())
    if "id" not in s:
        raise exceptions.AuthError("Invalid Authorization Bearer Token")

    return get_user_by_id(int(s["id"]))


def get_user_by_action_token(action, token):
    """
    Get the user by action token
    :param action: str
    :param token: str
    :return: AuthUser
    """
    data = utils.unsign_url_safe(token,
                                 secret_key=get_jwt_secret(),
                                 salt=action)
    if data is None:
        raise exceptions.AuthError("Invalid Token")
    return get_user_by_id(int(data))


def authenticate(username, password):
    """
    To authenticate a user with user and password
    :param username: 
    :param password: 
    :return: UserModel
    """
    user = models.AuthUser.get_by_username(username)
    return _user(user) if user and user.password_matched(password) else None


def authenticate_social_login(provider, social_id):
    """
    To authenticate with social login
    :param provider:
    :param social_id:
    :return: UserModel
    """
    user = models.AuthUserSocialLogin.get_user(provider, social_id)
    return _user(user) if user else None


def login_user(user):
    """
    Login and create a session
    :param user:
    :return:
    """
    def cb():
        if user:
            if __options__.get("require_email_verification") and not user.email_verified:
                raise exceptions.VerifyEmailError()
            user.update(last_login_at=utc_now())
            flask_login.login_user(UserModel(user))
            return user
        return None
    return signals.on_login(cb)


#
class UserModel(flask_login.UserMixin):

    def __init__(self, user):
        self.user = user.user if isinstance(user, self.__class__) else user
        self.user_salt = "USER:%s" % self.user.id

    def __getattr__(self, item):
        return getattr(self.user, item)

    # ------ FLASK-LOGIN REQUIRED METHODS ----------------------------------
    @property
    def is_active(self):
        return self.active

    # ---------- END FLASK-LOGIN REQUIREMENTS ------------------------------

    def change_username(self, username):
        """
        Change user's login email
        :param user: AuthUser
        :param email:
        :return:
        """
        def cb():

            if self.login_method == "username" and "@" in username:
                raise exceptions.AuthError(_("Username can't be an email"))
            elif self.login_method == "email" and "@" not in username:
                raise exceptions.AuthError(_("Invalid email login"))
            if "@" in username:
                if not utils.is_email_valid(username):
                    raise exceptions.AuthError("Email address invalid")
            elif not utils.is_username_valid(username):
                raise exceptions.AuthError("Username invalid")

            # Change both email and
            if self.login_method == "email":
                if not models.AuthUser.get_by_username(username) \
                        and not models.AuthUser.get_by_email(username):
                    self.user.change_username(username)
                    self.user.change_email(username)
            else:
                self.user.change_username(username)
            return username
        return signals.on_username_changed(cb)

    def change_email(self, email):
        """
        Change user's login email
        :param user: AuthUser
        :param email:
        :return:
        """

        def cb():
            if not utils.is_email_valid(email):
                raise exceptions.AuthError("Email address invalid")
            self.user.change_email(email)
            return email
        return signals.on_email_changed(cb)

    def update_info(self, **kwargs):
        """
        Change user's login email
        :param user:
        :param email:
        :return:
        """
        def cb():
            kwargs.pop("email", None)
            kwargs.pop("username", None)
            self.user.update(**kwargs)
            return kwargs
        return signals.on_info_updated(cb)

    def change_password(self, password):
        """
        Change a user's password
        :param user:
        :param password:
        :param password_confirm:
        :return:
        """

        def cb():
            if not utils.is_password_valid(password):
                raise exceptions.AuthError("Invalid Password")
            self.user.change_password(password)
            return True
        return signals.on_password_changed(cb)

    def reset_password(self):
        """
        Return the new random password that has been reset
        :param user_login: AuthUserLogin
        :return: string - the new password
        """

        def cb():
            password = get_random_password()
            self.change_password(password)
            return password
        return signals.on_password_reset(cb)

    def create_jwt(self, expires_in=None):
        """
        Create a secure timed JWT token that can be passed. It save the user id,
        which later will be used to retrieve the data

        :param user: AuthUser, the user's object
        :param expires_in: - time in second for the token to expire
        :return: string
        """
        s = utils.sign_jwt(data={"id": self.user.id},
                           secret_key=get_jwt_secret(),
                           salt=get_jwt_salt(),
                           expires_in=expires_in or get_jwt_ttl())
        return s

    def create_action_token(self, action, expires_in):
        """
        Create a url safe action token attached to the user
        :param action:
        :param expires_in:
        :return:
        """
        return utils.sign_url_safe(self.user.id,
                                   secret_key=get_jwt_secret(),
                                   salt=action,
                                   expires_in=expires_in)

    def sign_data(self, data, expires_in=None, url_safe=True):
        """
        To safely sign a user data. It will be signed with the user key
        :param data: mixed
        :param expires_in: The time for it to expire
        :param url_safe: bool. If true it will allow it to be passed in URL
        :return: str -  the token/signed data
        """
        if url_safe:
            return utils.sign_url_safe(data,
                                       secret_key=self.secret_key,
                                       salt=self.user_salt,
                                       expires_in=expires_in)
        else:
            return utils.sign_data(data,
                                   secret_key=self.secret_key,
                                   salt=self.user_salt,
                                   expires_in=expires_in)

    def unsign_data(self, data, url_safe=True):
        """
        Retrieve the signed data. If it is expired, it will throw an exception
        :param data: token/signed data
        :param url_safe: bool. If true it will allow it to be passed in URL
        :return: mixed, the data in its original form
        """
        if url_safe:
            return utils.unsign_url_safe(data,
                                         secret_key=self.secret_key,
                                         salt=self.user_salt)
        else:
            return utils.unsign_data(data,
                                     secret_key=self.secret_key,
                                     salt=self.user_salt)

    def signed_data_match(self, data, matched_data, url_safe=True):
        """
        See if a data matched a signed one
        :param data:
        :param matched_data:
        :param url_safe:
        :return:
        """
        try:
            u_data = self.unsign_data(data, url_safe=url_safe)
            return u_data == matched_data
        except Exception as e:
            return False

    def send_email(self, template, **kwargs):
        """
        To send email to user
        :param template:
        :param kwargs:
        :return:
        """
        user_data = {
            "id": self.id,
            "username": self.username,
            "name": self.name,
            "first_name": self.first_name,
            "last_name": self.last_name,
            "email": self.email
        }
        kwargs.pop("user", None)
        send_mail(to=self.email, template=template, user=user_data, **kwargs)

    def send_password_reset(self, base_url=None, view_class=None, **kw):
        """
        Reset a password and send email
        :param user: AuthUser
        :param email: str - The auth user login email
        :param base_url: str - By default it will use the current url, base_url will allow custom url
        :param template: str - The email template
        :param method: str - token or email - The method to reset the password
        :param view_class: obj - The view instance of the login
        :param kwargs: Any data to pass
        :return:
        """

        view = view_class or views.auth.Login
        endpoint_reset = getattr(view, "reset_password")
        endpoint_login = getattr(view, "login")
        action = "reset-password"

        method = __options__.get("reset_password_method", "TOKEN")
        template = __options__.get("email_templates.reset_password",
                                   "auth/reset-password.txt")
        new_password = None

        if method.upper() == "TOKEN":
            expires_in = __options__.get("reset_password_token_ttl", 1)
            action_token = self.create_action_token(action, expires_in)
            signed_data = self.sign_data(action, expires_in=expires_in)
            url = _url_for_email(endpoint_reset,
                                 base_url=base_url,
                                 action_token=action_token,
                                 signed_data=signed_data)
        else:
            new_password = self.reset_password()
            url = _url_for_email(endpoint_login, base_url=base_url)

        self.send_email(template=template,
                        action={
                            "reset_method": method.upper(),
                            "url": url,
                            "new_password": new_password
                        },
                        data=kw)

    def send_verification_email(self, base_url=None, view_class=None, **kw):

        template = __options__.get("email_templates.verify_email",
                                   "auth/verify-email.txt")

        url = self._create_verify_email_token_url(base_url=base_url,
                                                  view_class=view_class)

        self.send_email(template=template,
                        action={
                            "url": url,
                        },
                        data=kw)

    def send_welcome_email(self, base_url=None, view_class=None, **kw):

        verify_email = __options__.get("require_email_verification") or False
        template = __options__.get("email_templates.welcome",
                                   "auth/welcome.txt")

        url = self._create_verify_email_token_url(base_url=base_url,
                                                  view_class=view_class)

        self.send_email(template=template,
                        action={
                            "url": url,
                            "require_email_verification": verify_email,
                        },
                        data=kw)

    def _create_verify_email_token_url(self, base_url=None, view_class=None):
        """
        To create a verify email token url
        :param user: (object) AuthUser
        :param base_url: a base_url to use instead of the native one
        :param view_class: (obj) the view class, to allow build the url
        :return: string
        """
        view = view_class or views.auth.Login
        endpoint = getattr(view, "verify_email")
        action = "verify-email"
        expires_in = __options__.get("verify_email_token_ttl") or (60 * 24)
        action_token = self.create_action_token(action, expires_in)
        signed_data = self.sign_data(action, expires_in=expires_in)
        url = _url_for_email(endpoint,
                             base_url=base_url,
                             action_token=action_token,
                             signed_data=signed_data)

        return url

    def add_social_login(self, provider, social_id):
        """
        Add social login to the current user
        :param provider:
        :param social_id:
        :return:
        """
        models.AuthUserSocialLogin.new(user=self,
                                       provider=provider,
                                       social_id=social_id)

# ------------------------------------------------------------------------------
# EMAIL SENDING


def _url_for_email(endpoint, base_url=None, **kw):
    """
    Create an external url_for by using a custom base_url different from the domain we
    are on
    :param endpoint:
    :param base_url:
    :param kw:
    :return:
    """
    base_url = base_url or get_config("MAIL_EXTERNAL_BASE_URL")
    _external = True if not base_url else False
    url = url_for(endpoint, _external=_external, **kw)
    if base_url and not _external:
        url = "%s/%s" % (base_url.strip("/"), url.lstrip("/"))
    return url


def session_set_require_password_change(change=True):
    session["auth:require_password_change"] = change


def session_get_require_password_change():
    return session.get("auth:require_password_change")


# ------------------------------------------------------------------------------
# CLI
from harambe.cli import CLI


class AuthCLI(CLI):
    def __init__(self, command, click):
        @command("auth:create-super-admin")
        @click.argument("email")
        def create_super_admin(email):
            """
            To create a super admin by providing the email address
            """
            print("-" * 80)
            print("Harambe Auth: Create Super Admin")
            print("Email: %s" % email)
            try:
                password = get_random_password()
                user = register(email=email, password=password,
                                name="SuperAdmin", role_name="Superadmin")
                user.update(require_password_change=True)

                print("Password: %s" % password)
            except Exception as e:
                print("ERROR: %s" % e)

            print("Done!")

        @command("auth:reset-password")
        @click.argument("email")
        def reset_password(email):
            """
            To reset password by email
            """
            print("-" * 80)
            print("Harambe Auth: Reset Password")
            try:
                ul = models.AuthUserLogin.get_by_email(email)

                if not ul:
                    raise Exception("Email '%s' doesn't exist" % email)
                print(ul.email)
                password = get_random_password()
                ul.change_password(password)
                ul.update(require_password_change=True)
                print("Email: %s" % email)
                print("New Password: %s" % password)
            except Exception as e:
                print("ERROR: %s" % e)

            print("Done!")

        @command("auth:user-info")
        @click.option("--email")
        @click.option("--id")
        def reset_password(email=None, id=None):
            """
            Get the user info by email or ID
            """
            print("-" * 80)
            print("Harambe Auth: User Info")
            print("")
            try:
                if email:
                    ul = models.AuthUserLogin.get_by_email(email)
                    if not ul:
                        raise Exception("Invalid Email address")
                    user_info = ul.user
                elif id:
                    user_info = models.AuthUser.get(id)
                    if not user_info:
                        raise Exception("Invalid User ID")

                k = [
                    ("ID", "id"), ("Name", "name"),
                    ("First Name", "first_name"),
                    ("Last Name", "last_name"), ("Signup", "created_at"),
                    ("Last Login", "last_login"),
                    ("Signup Method", "register_method"),
                    ("Is Active", "is_active")
                ]
                print("Email: %s" % user_info.get_email_login().email)
                for _ in k:
                    print("%s : %s" % (_[0], getattr(user_info, _[1])))

            except Exception as e:
                print("ERROR: %s" % e)

            print("")
            print("Done!")


# ---

from .decorators import *





import datetime
import arrow
from . import exceptions
from harambe import (db,
                   utils,
                   send_mail,
                   _)
from harambe.exceptions import ModelError


class AuthUserRole(db.Model):

    SUPERADMIN = "SUPERADMIN"  # ALL MIGHTY, RESERVED FOR SYS ADMIN
    ADMIN = "ADMIN"  # App/Site admin
    MANAGER = "MANAGER"  # Limited access, but can approve EDITOR Data
    EDITOR = "EDITOR"  # Rights to write, manage, publish own data
    CONTRIBUTOR = "CONTRIBUTOR"  # Rights to only write and read own data
    MODERATOR = "MODERATOR"   # Moderate content
    MEMBER = "MEMBER"  # Just a member

    # BASIC ROLES
    ROLES = [(99, SUPERADMIN),
               (89, ADMIN),
               (59, MANAGER),
               (49, EDITOR),
               (39, CONTRIBUTOR),
               (29, MODERATOR),
               (10, MEMBER)]

    name = db.Column(db.String(75), index=True)
    level = db.Column(db.Integer, index=True)

    @classmethod
    def _syncdb(cls):
        """
        Harambe specific
        To setup some models data after
        :return:
        """
        [cls.new(level=r[0], name=r[1]) for r in cls.ROLES]

    @classmethod
    def new(cls, name, level):
        name = cls.slug_name(name)
        if not cls.get_by_name(name) and not cls.get_by_level(level):
            return cls.create(name=name, level=level)

    @classmethod
    def get_by_name(cls, name):
        name = cls.slug_name(name)
        return cls.query().filter(cls.name == name).first()

    @classmethod
    def get_by_level(cls, level):
        return cls.query().filter(cls.level == level).first()

    @classmethod
    def slug_name(cls, name):
        return utils.slugify(name)


class AuthUser(db.Model):
    """
    User
    """
    role_id = db.Column(db.Integer, db.ForeignKey(AuthUserRole.id))
    login_method = db.Column(db.String(75))
    username = db.Column(db.String(255), index=True, unique=True)
    email = db.Column(db.EmailType, index=True)
    email_verified = db.Column(db.Boolean, default=False)
    password_hash = db.Column(db.String(255))
    require_password_change = db.Column(db.Boolean, default=False)
    first_name = db.Column(db.String(255))
    last_name = db.Column(db.String(255))
    profile_image = db.Column(db.StorageObjectType)
    options = db.Column(db.JSONType)
    active = db.Column(db.Boolean, default=True, index=True)
    last_login_at = db.Column(db.DateTime)
    locale = db.Column(db.String(10), default="en")
    timezone = db.Column(db.String(255))
    # A secret key to sign data per user. If a key is compromised,
    # we can reset this one only and not affect anyone else
    secret_key = db.Column(db.String(255))
    # Relationship to role
    role = db.relationship(AuthUserRole)

    @classmethod
    def encrypt_password(cls, password):
        return utils.encrypt_string(password)

    @classmethod
    def new(cls,
            username,
            password,
            email=None,
            first_name=None,
            last_name=None,
            login_method=None,
            role="MEMBER"
            ):
        """
        Create a new user
        :param username: str
        :param password: str
        :param email: str
        :param first_name: str
        :param last_name: str
        :param login_method: str
        :param role: str
        :return: AuthUser
        """
        data = {
            "first_name": first_name,
            "last_name": last_name,
            "email": email
        }

        username = username.strip().lower()
        if "@" in username and not email:
            if not utils.is_email_valid(username):
                exceptions.AuthError(_("Invalid username"))
            data["email"] = username
        elif email:
            if not utils.is_email_valid(email):
                exceptions.AuthError(_("Invalid username"))

        if not utils.is_username_valid(username):
            exceptions.AuthError(_("Invalid username"))
        if not utils.is_password_valid(password):
            raise exceptions.AuthError(_("Password is invalid"))

        if cls.get_by_username(username):
            raise exceptions.AuthError(_("Username exists already"))
        _email = data.get("email")
        if _email:
            if cls.get_by_email(_email):
                raise exceptions.AuthError(_("Email address exists already"))

        role = AuthUserRole.get_by_name(role or "MEMBER")
        if not role:
            raise exceptions.AuthError(_("Invalid user role"))

        data.update({
            "username": username,
            "password_hash": cls.encrypt_password(password),
            "email_verified": False,
            "login_method": login_method,
            "role": role
        })
        user = cls.create(**data)
        user.reset_secret_key()
        return user


    @classmethod
    def get_by_username(cls, username):
        """
        Return a User by email address
        """
        return cls.query().filter(cls.username == username).first()

    @classmethod
    def get_by_email(cls, email):
        """
        Return a User by email address
        """
        return cls.query().filter(cls.email == email).first()

    @classmethod
    def search_by_name(cls, query, name):
        """
        Make a search
        :param query:
        :param name:
        :return:
        """
        query = query.filter(db.or_(cls.first_name.contains(name),
                                    cls.last_name.contains(name)))
        return query

    @property
    def name(self):
        return self.first_name

    @property
    def full_name(self):
        return "%s %s" % (self.first_name, self.last_name)

    def change_username(self, username):
        """
        Change username
        :param username: email or str
        :return:
        """
        username = username.lower()
        if self.username != username:
            if self.get_by_username(username):
                raise exceptions.AuthError("Username exists already")
            self.update(username=username)

    def change_email(self, email, as_username=False):
        """
        Change account email
        :param email:
        :param as_username
        :return: the email provided
        """
        email = email.lower()
        data = {"email": email}
        if self.email != email:
            if self.get_by_email(email):
                raise exceptions.AuthError("Email exists already")
            if as_username:
                if self.username != email:
                    if self.get_by_username(email):
                        raise exceptions.AuthError("Username exists already")
                data["username"] = email

            self.update(**data)

    def change_password(self, password):
        """
        Change the password.
        :param password:
        :return:
        """
        self.update(password_hash=self.encrypt_password(password),
                    require_password_change=False)

        # Whenever the password is changed, reset the secret key to invalidate
        # any tokens in the wild
        self.reset_secret_key()

    def set_email_verified(self, verified=False):
        self.update(email_verified=verified)
        self.reset_secret_key()

    def password_matched(self, password):
        """
        Check if the password matched the hash
        :returns bool:
        """
        return utils.verify_encrypted_string(password, self.password_hash)

    def reset_secret_key(self):
        """
        Run this whenever there is a security change in the account.
        ie: password change.
        BTW, AuthUserLogin.change_password, already performs this method
        """
        self.update(secret_key=utils.guid())

    def set_options(self, **kwargs):
        """
        Set the options. Existing value will persist
        :param kwargs:
        :return:
        """
        options = self.options
        options.update(kwargs)
        self.update(options=options)

    def set_role(self, role):
        role_ = AuthUserRole.get_by_name(role.upper())
        if not role_:
            raise ModelError("Invalid user role: '%s'" % role)
        self.update(role=role_)

    def has_any_roles(self, *roles):
        """
        Check if user has any of the roles requested
        :param roles: tuple of roles string
        :return: bool
        """
        roles = map(utils.slugify, list(roles))

        return True \
            if AuthUserRole.query() \
            .join(AuthUser) \
            .filter(AuthUserRole.name.in_(roles)) \
            .filter(AuthUser.id == self.id) \
            .count() \
            else False


class AuthUserSocialLogin(db.Model):
    """
    Only store the provider name and social_id, to match user
    """
    user_id = db.Column(db.Integer, db.ForeignKey(AuthUser.id))
    provider = db.Column(db.String(50), index=True)
    social_id = db.Column(db.String(255), index=True)
    user = db.relationship(AuthUser, backref='social_logins')
    active = db.Column(db.Boolean, default=True)

    @classmethod
    def get_user(cls, provider, social_id):
        login = cls.query() \
            .filter(cls.provider == provider) \
            .filter(cls.social_id == social_id) \
            .first()
        return login.user if login else None

    @classmethod
    def new(cls, user, provider, social_id):
        """
        Create a new login
        :param user: AuthUser
        :param provider: str - ie: facebook, twitter, ...
        :param social_id: str - an id associated to provider
        :return:
        """
        if cls.get_user(provider, social_id):
            raise exceptions.AuthError("Social Login exists already")

        return cls.create(user_id=user.id,
                          provider=provider,
                          social_id=social_id)





"""
Extensions
"""

import re
import logging
from flask import request, current_app
import flask_cloudy
import flask_recaptcha
import flask_seasurf
import flask_kvsession
import flask_caching
import ses_mailer
import flask_mail
import flask_s3
import flask_babel
from flask_babel import lazy_gettext as _
from . import Harambe, init_app, utils, exceptions, g
from.extras import jinja_helpers
from paginator import Paginator

__all__ = ["cache",
           "storage",
           "recaptcha",
           "csrf",
           "send_mail",
           "paginate",
           "_",
           ]


def _setup(app):

    check_config_keys = [
        "SECRET_KEY",
        "ADMIN_EMAIL",
        "CONTACT_EMAIL",
        "MAIL_URI",
        "MAIL_SENDER",
        "DB_URI",
        "DATETIME_TIMEZONE",
        "DATETIME_FORMAT",
        "RECAPTCHA_SITE_KEY",
        "RECAPTCHA_SECRET_KEY"
    ]
    for k in check_config_keys:
        if k not in app.config \
                or not app.config.get(k):
            msg = "Missing config key value: %s " % k
            logging.warning(msg)


# ------------------------------------------------------------------------------

# Session
#
# It uses KV session to allow multiple backend for the session
def _session(app):
    store = None
    uri = app.config.get("SESSION_URI")
    if uri:
        parse_uri = utils.urlparse(uri)
        scheme = parse_uri.scheme
        username = parse_uri.username
        password = parse_uri.password
        hostname = parse_uri.hostname
        port = parse_uri.port
        bucket = parse_uri.path.strip("/")

        if "redis" in scheme:
            import redis
            from simplekv.memory.redisstore import RedisStore
            conn = redis.StrictRedis.from_url(url=uri)
            store = RedisStore(conn)
        elif "s3" in scheme or "google_storage" in scheme:
            from simplekv.net.botostore import BotoStore
            import boto
            if "s3" in scheme:
                _con_fn = boto.connect_s3
            else:
                _con_fn = boto.connect_gs
            conn = _con_fn(username, password)
            _bucket = conn.create_bucket(bucket)
            store = BotoStore(_bucket)
        elif "memcache" in scheme:
            import memcache
            from simplekv.memory.memcachestore import MemcacheStore
            host_port = "%s:%s" % (hostname, port)
            conn = memcache.Client(servers=[host_port])
            store = MemcacheStore(conn)
        elif "sql" in scheme:
            from simplekv.db.sql import SQLAlchemyStore
            from sqlalchemy import create_engine, MetaData
            engine = create_engine(uri)
            metadata = MetaData(bind=engine)
            store = SQLAlchemyStore(engine, metadata, 'kvstore')
            metadata.create_all()
        else:
            raise exceptions.HarambeError("Invalid Session Store. '%s' provided" % scheme)
    if store:
        flask_kvsession.KVSessionExtension(store, app)

# ------------------------------------------------------------------------------
# Mailer
class _Mailer(object):
    """
    config key: MAIL_*
    A simple wrapper to switch between SES-Mailer and Flask-Mail based on config
    """
    mail = None
    provider = None
    config = None
    _template = None

    @property
    def validated(self):
        return bool(self.mail)

    def init_app(self, app):
        self.config = app.config
        scheme = None

        mailer_uri = self.config.get("MAIL_URI")
        if mailer_uri:
            mailer_uri = utils.urlparse(mailer_uri)
            scheme = mailer_uri.scheme
            hostname = mailer_uri.hostname

            # Using ses-mailer
            if "ses" in scheme.lower():
                self.provider = "SES"

                access_key = mailer_uri.username or app.config.get("AWS_ACCESS_KEY_ID")
                secret_key = mailer_uri.password or app.config.get("AWS_SECRET_ACCESS_KEY")
                region = hostname or self.config.get("AWS_REGION", "us-east-1")
                
                self.mail = ses_mailer.Mail(aws_access_key_id=access_key,
                                            aws_secret_access_key=secret_key,
                                            region=region,
                                            sender=self.config.get("MAIL_SENDER"),
                                            reply_to=self.config.get("MAIL_REPLY_TO"),
                                            template=self.config.get("MAIL_TEMPLATE"),
                                            template_context=self.config.get("MAIL_TEMPLATE_CONTEXT"))

            # SMTP will use flask-mail
            elif "smtp" in scheme.lower():
                self.provider = "SMTP"

                class _App(object):
                    config = {
                        "MAIL_SERVER": mailer_uri.hostname,
                        "MAIL_USERNAME": mailer_uri.username,
                        "MAIL_PASSWORD": mailer_uri.password,
                        "MAIL_PORT": mailer_uri.port,
                        "MAIL_USE_TLS": True if "tls" in mailer_uri.scheme else False,
                        "MAIL_USE_SSL": True if "ssl" in mailer_uri.scheme else False,
                        "MAIL_DEFAULT_SENDER": app.config.get("MAIL_SENDER"),
                        "TESTING": app.config.get("TESTING"),
                        "DEBUG": app.config.get("DEBUG")
                    }
                    debug = app.config.get("DEBUG")
                    testing = app.config.get("TESTING")

                _app = _App()
                self.mail = flask_mail.Mail(app=_app)

                _ses_mailer = ses_mailer.Mail(template=self.config.get("MAIL_TEMPLATE"),
                                              template_context=self.config.get("MAIL_TEMPLATE_CONTEXT"))
                self._template = _ses_mailer.parse_template
            else:
                logging.warning("Mailer Error. Invalid scheme '%s'" % scheme)

    def send(self, to, subject=None, body=None, reply_to=None, template=None, **kwargs):
        """
        To send email
        :param to: the recipients, list or string
        :param subject: the subject
        :param body: the body
        :param reply_to: reply_to
        :param template: template, will use the templates instead
        :param kwargs: context args
        :return: bool - True if everything is ok
        """
        sender = self.config.get("MAIL_SENDER")
        recipients = [to] if not isinstance(to, list) else to
        kwargs.update({
            "subject": subject,
            "body": body,
            "reply_to": reply_to
        })

        if not self.validated:
            raise exceptions.HarambeError("Mail configuration error")

        if self.provider == "SES":
            kwargs["to"] = recipients
            if template:
                self.mail.send_template(template=template, **kwargs)
            else:
               self.mail.send(**kwargs)

        elif self.provider == "SMTP":
            if template:
                data = self._template(template=template, **kwargs)
                kwargs["subject"] = data["subject"]
                kwargs["body"] = data["body"]
            kwargs["recipients"] = recipients
            kwargs["sender"] = sender

            # Remove invalid Messages keys
            _safe_keys = ["recipients", "subject", "body", "html", "alts",
                          "cc", "bcc", "attachments", "reply_to", "sender",
                           "date", "charset", "extra_headers", "mail_options",
                           "rcpt_options"]
            for k in kwargs.copy():
                if k not in _safe_keys:
                    del kwargs[k]

            message = flask_mail.Message(**kwargs)
            self.mail.send(message)
        else:
            raise exceptions.HarambeError("Invalid mail provider. Must be 'SES' or 'SMTP'")

# ------------------------------------------------------------------------------

# Assets Delivery
class _AssetsDelivery(flask_s3.FlaskS3):
    def init_app(self, app):
        delivery_method = app.config.get("ASSETS_DELIVERY_METHOD")
        if delivery_method and delivery_method.upper() in ["S3", "CDN"]:
            #with app.app_context():
            is_secure = False #request.is_secure

            if delivery_method.upper() == "CDN":
                domain = app.config.get("ASSETS_DELIVERY_DOMAIN")
                if "://" in domain:
                    domain_parsed = utils.urlparse(domain)
                    is_secure = domain_parsed.scheme == "https"
                    domain = domain_parsed.netloc
                app.config.setdefault("S3_CDN_DOMAIN", domain)

            app.config["FLASK_ASSETS_USE_S3"] = True
            app.config["FLASKS3_ACTIVE"] = True
            app.config["FLASKS3_URL_STYLE"] = "path"
            app.config.setdefault("FLASKS3_USE_HTTPS", is_secure)
            app.config.setdefault("FLASKS3_ONLY_MODIFIED", True)
            app.config.setdefault("FLASKS3_GZIP", True)
            app.config.setdefault("FLASKS3_BUCKET_NAME", app.config.get("AWS_S3_BUCKET_NAME"))

            super(self.__class__, self).init_app(app)

# ------------------------------------------------------------------------------

# Set CORS
def set_cors_config(app):
    """
    Flask-Core (3.x.x) extension set the config as CORS_*,
     But we'll hold the config in CORS key.
     This function will convert them to CORS_* values
    :param app:
    :return:
    """
    if "CORS" in app.config:
        for k, v in app.config["CORS"].items():
            _ = "CORS_" + k.upper()
            if _ not in app.config:
                app.config[_] = v

# ------------------------------------------------------------------------------


def send_mail(to, **kwargs):
    """
    Alias to mail.send()
    :param to:
    :param kwargs:
    :return:
    """
    return mail.send(to=to, **kwargs)

init_app(_setup)

init_app(_session)

# CORS
init_app(set_cors_config)

# Mail
mail = _Mailer()
init_app(mail.init_app)

# Cache
cache = flask_caching.Cache()
init_app(cache.init_app)

# Storage
storage = flask_cloudy.Storage()
init_app(storage.init_app)

# Recaptcha
recaptcha = flask_recaptcha.ReCaptcha()
init_app(recaptcha.init_app)

# CSRF
csrf = flask_seasurf.SeaSurf()
init_app(csrf.init_app)

# Assets delivery
assets_delivery = _AssetsDelivery()
init_app(assets_delivery.init_app)


class _Paginator(object):
    def init_app(self, app):
        self.per_page = app.config.get("PAGINATION_PER_PAGE", 20)

    def paginate(self, iter, **kwargs):
        if "page" not in kwargs:
            kwargs["page"] = int(request.args.get('page', 1))
        return Paginator(iter, **kwargs)

_paginator = _Paginator()

paginate = _paginator.paginate


# Babel
babel = flask_babel.Babel()


def init_babel(app):
    babel.init_app(app)

    # languages = app.config.get("LANGUAGES", {})
    #
    # @babel.localeselector
    # def get_locale():
    #     lang = request.params.get("lang", "en")
    #     return request.accept_languages.best_match(languages.keys())
init_app(init_babel)


# ------------------------------------------------------------------------------




from harambe import Harambe
from harambe.contrib.app_option import AppOption

__options__ = {}

class Main(Harambe):

    NS = "MaintenancePage"

    app_option = AppOption(__name__)

    @classmethod
    def _register(cls, app, **kwargs):
        template = __options__.get("template", "maintenance_page/Main/index.jade")
        super(cls, cls)._register(app, **kwargs)

        cls.app_option.init({
            "status": False,
            "exclude": []
        }, "Maintenance Page Option")

        @app.before_request
        def on_maintenance():
            if cls.app_option.get("status") is True:
                return cls.render(_layout=template), 503





"""
Harambe

Harambe is a Flask based framework to help quickly develop web applications, by
adding structure to your views and templates.

Philosophy:

To create a framework that runs everywhere, regardless of the platform, by
providing cloud balh...

It made the following decisions for you: (of course you can change them)


It comes with pre-built components:


And it is still Flask.

https://github.com/mardix/harambe

"""

import os
from setuptools import setup, find_packages

base_dir = os.path.dirname(__file__)

__about__ = {}
with open(os.path.join(base_dir, "harambe", "__about__.py")) as f:
    exec(f.read(), __about__)

with open('requirements.txt') as f:
    install_requires = f.read().splitlines()

setup(
    name=__about__["__title__"],
    version=__about__["__version__"],
    license=__about__["__license__"],
    author=__about__["__author__"],
    author_email=__about__["__email__"],
    description=__about__["__summary__"],
    url=__about__["__uri__"],
    long_description=__doc__,
    py_modules=['harambe'],
    entry_points=dict(console_scripts=[
        'goham=harambe.cli:cmd',
        'harambe=harambe.cli:cmd',

    ]),
    include_package_data=True,
    packages=find_packages(exclude=["*.tests", "*.tests.*", "tests.*", "tests"]),
    install_requires=install_requires,
    keywords=['flask',
              'Harambe',
              'templates',
              'views',
              'classy',
              'framework',
              "mvc",
              "blueprint",
              "harambe"],
    platforms='any',
    classifiers=[
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ],
    zip_safe=False
)




"""
Error Page

This plugin to display customize error page

Can be called as standalone
"""
from __future__ import division
import logging
from harambe import (Harambe,
                   page_meta,
                   register_package,
                   abort)

from harambe import exceptions
from sqlalchemy.exc import SQLAlchemyError


__version__ = "1.0.0"
__options__ = {}

# A callback function to use as renderer instead of the built in one
# A use case can be
renderer = None

"""
Example:

@render_as_json
def my_error(e):
    return {
        "error": True,
        "code": e.code,
        "message": e.description
    }, e.code
renderer = my_error
"""



class Main(Harambe):

    @classmethod
    def _register(cls, app, **kwargs):

        template_page = __options__.get("template", "error_page/Main/index.jade")

        super(cls, cls)._register(app, **kwargs)

        @app.errorhandler(400)
        @app.errorhandler(401)
        @app.errorhandler(403)
        @app.errorhandler(404)
        @app.errorhandler(405)
        @app.errorhandler(406)
        @app.errorhandler(408)
        @app.errorhandler(409)
        @app.errorhandler(410)
        @app.errorhandler(413)
        @app.errorhandler(414)
        @app.errorhandler(429)
        @app.errorhandler(500)
        @app.errorhandler(501)
        @app.errorhandler(502)
        @app.errorhandler(503)
        @app.errorhandler(504)
        @app.errorhandler(505)
        def index(error):

            #page_meta("Error %s" % error.code)


            # if isinstance(error, SQLAlchemyError):
            #     error = SQLAlchemyHTTPException(error)

            # we'll log non 4** errors
            if int(error.code // 100) != 4:
                _error = str(error)
                _error += " - HTTException Code: %s" % error.code
                _error += " - HTTException Description: %s" % error.description
                logging.error(_error)

            if renderer:
                return renderer(error)
            else:
                return cls.render(error=error), error.code




"""
CONFIGURATION

Class based config file, where each class is an environment:
ie: Dev = Development, Prod=Production, ...

Harambe: Global Configuration

Global config shared by all applications

The method allow multiple configuration

By default, it is expecting the Dev and Prod, but you can add your class
which extends from BaseConfig

- Access in templates
You have the ability to access these config in your template
just use the global variable `config`
ie:
    {{ config.APPLICATION_NAME }}

"""

from . import (get_var_path,
               ROOT_DIR,
               VAR_DIR,
               AVAILABLE_APPS)


class BaseConfig(object):
    """
    Base Configuration.
    """


# ------------------------------------------------------------------------------
#: Config

    #: Site's name or Name of the application
    APPLICATION_NAME = "Harambe"

    #: The application url
    APPLICATION_URL = ""

    #: Version of application
    APPLICATION_VERSION = "0.0.1"

    #: Google Analytics ID
    GOOGLE_ANALYTICS_ID = ""

# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------
#: MISC CONFIG

    #: Required to setup. This email will have SUPER USER role
    ADMIN_EMAIL = None

    #: The address to receive email when using the contact page
    CONTACT_EMAIL = None

    #: PAGINATION_PER_PAGE : Total entries to display per page
    PAGINATION_PER_PAGE = 20

    # MAX_CONTENT_LENGTH
    # If set to a value in bytes, Flask will reject incoming requests with a
    # content length greater than this by returning a 413 status code
    MAX_CONTENT_LENGTH = 2 * 1024 * 1024

    # To remove whitespace off the HTML result
    COMPRESS_HTML = False

# ------------------------------------------------------------------------------
#: DATETIME TIMEZONE + FORMAT

    #: DATE AND TIME FORMAT
    # By default, Harambe uses Arrow
    # http://crsmithdev.com/arrow/
    # https://github.com/crsmithdev/arrow

    # Timezone
    DATETIME_TIMEZONE = "US/Eastern"

    # Datetime format
    # Dict of all dates format used by your application, inn your template
    # {{ date_var | local_datetime }}
    # or {{ date_var | local_datetime('long_datetime') }}
    # or {{ date_var | local_datetime('MM/DD/YYYY') }}
    # By default it will fallback to default, or use the format provided
    # For all the tokens: http://crsmithdev.com/arrow/#tokens
    # You can add as many keys as you want
    DATETIME_FORMAT = {
        "default": "MM/DD/YYYY",
        "date": "MM/DD/YYYY",
        "datetime": "MM/DD/YYYY hh:mm a",
        "time": "hh:mm a",
        "long_datetime": "dddd, MMMM D, YYYY hh:mm a",
    }

# ------------------------------------------------------------------------------
#: AWS CREDENTIALS

    #: AWS Credentials
    # AWS is used by lots of extensions
    # For: S3, SES Mailer, flask S3.

    # The AWS Access KEY
    AWS_ACCESS_KEY_ID = ""

    # Secret Key
    AWS_SECRET_ACCESS_KEY = ""

    # The bucket name for S3
    AWS_S3_BUCKET_NAME = ""

    # The default region name
    AWS_REGION_NAME = "us-east-1"

# ------------------------------------------------------------------------------
#: DATABASES

    #: DB_URI
    #: format: engine://USERNAME:PASSWORD@HOST:PORT/DB_NAME
    DB_URI = "sqlite:////%s/db.db" % VAR_DIR

    #: REDIS_URI
    #: format: USERNAME:PASSWORD@HOST:PORT
    REDIS_URI = None

# ------------------------------------------------------------------------------
#: ASSETS DELIVERY

    # ASSETS DELIVERY allows to serve static files from S3, Cloudfront or other CDN

    # The delivery method:
    #   - None: will use the local static files
    #   - S3: Will use AWS S3. By default it will use the bucket name set in AWS_S3_BUCKET_NAME
    #       When S3, AWS_ACCESS_KEY_ID and AWS_SECRET_ACCESS_KEY are required to upload files
    #   - CDN: To use a CDN. ASSETS_DELIVERY_DOMAIN need to have the CDN domain
    ASSETS_DELIVERY_METHOD = None

    # Set the base domain of the CDN
    ASSETS_DELIVERY_DOMAIN = None


# ------------------------------------------------------------------------------
#: SESSION

    #: SESSION
    #: Flask-KVSession is used to save the user's session
    #: Set the SESSION_URI by using these examples below to set KVSession
    #: To use local session, just set SESSION_URI to None
    #:
    #: Redis: redis://username:password@host:6379/db
    #: S3: s3://username:password@s3.aws.amazon.com/bucket
    #: Google Storage: google_storage://username:password@cloud.google.com/bucket
    #: SQL: postgresql://username:password@host:3306/db
    #:      mysql+pysql://username:password@host:3306/db
    #:      sqlite://
    #: Memcached: memcache://host:port
    #:
    SESSION_URI = None

# ------------------------------------------------------------------------------
#: STORAGE

    #: STORAGE
    #: Flask-Cloudy is used to save upload on S3, Google Storage,
    #: Cloudfiles, Azure Blobs, and Local storage
    #: When using local storage, they can be accessed via http://yoursite/files
    #:

    #: STORAGE_PROVIDER:
    # The provider to use. By default it's 'LOCAL'.
    # You can use:
    # LOCAL, S3, GOOGLE_STORAGE, AZURE_BLOBS, CLOUDFILES
    STORAGE_PROVIDER = "LOCAL"

    #: STORAGE_KEY
    # The storage key. Leave it blank if PROVIDER is LOCAL
    STORAGE_KEY = AWS_ACCESS_KEY_ID

    #: STORAGE_SECRET
    #: The storage secret key. Leave it blank if PROVIDER is LOCAL
    STORAGE_SECRET = AWS_SECRET_ACCESS_KEY

    #: STORAGE_REGION_NAME
    #: The region for the storage. Leave it blank if PROVIDER is LOCAL
    STORAGE_REGION_NAME = AWS_REGION_NAME

    #: STORAGE_CONTAINER
    #: The Bucket name (for S3, Google storage, Azure, cloudfile)
    #: or the directory name (LOCAL) to access
    STORAGE_CONTAINER = get_var_path("uploads")

    #: STORAGE_SERVER
    #: Bool, to serve local file
    STORAGE_SERVER = True

    #: STORAGE_SERVER_URL
    #: The url suffix for local storage
    STORAGE_SERVER_URL = "files"

# ------------------------------------------------------------------------------
#: MAIL

    # AWS SES
    # To use AWS SES to send email
    #:
    #: - To use the default AWS credentials (AWS_ACCESS_KEY_ID, AWS_SECRET_ACCESS_KEY)
    #: set MAIL_URI = "ses://"
    #: * To use a different credential:
    #: set MAIL_URI = "ses://{access_key}:{secret_key}@{region}"
    #:
    #: *** uncomment if you are using SMTP instead
    # MAIL_URI = "ses://"

    # SMTP
    #: If you are using SMTP, it will use Flask-Mail
    #: The uri for the smtp connection. It will use Flask-Mail
    #: format: smtp://USERNAME:PASSWORD@HOST:PORT
    #: with sll -> smtp+ssl://USERNAME:PASSWORD@HOST:PORT
    #: with ssl and tls -> smtp+ssl+tls://USERNAME:PASSWORD@HOST:PORT
    #:
    #: *** comment out if you are using SES instead
    # MAIL_URI = "smtp+ssl://{username}:{password}@{host}:{port}"\
    #    .format(username="", password="", host="smtp.gmail.com", port=465)

    #: MAIL_SENDER - The sender of the email by default
    #: For SES, this email must be authorized
    MAIL_SENDER = ADMIN_EMAIL

    #: MAIL_REPLY_TO
    #: The email to reply to by default
    MAIL_REPLY_TO = ADMIN_EMAIL

    #: MAIL_TEMPLATE
    #: a directory that contains the email template or a dict
    MAIL_TEMPLATE = get_var_path("mail-templates")

    #: MAIL_TEMPLATE_CONTEXT
    #: a dict of all context to pass to the email by default
    MAIL_TEMPLATE_CONTEXT = {
        "params": {
            "site_name": APPLICATION_NAME,
            "site_url": APPLICATION_URL
        }
    }

# ------------------------------------------------------------------------------
#: CACHE

    #: Flask-Cache is used to caching

    #: CACHE_TYPE
    #: The type of cache to use
    #: null, simple, redis, filesystem,
    CACHE_TYPE = "simple"

    #: CACHE_REDIS_URL
    #: If CHACHE_TYPE is 'redis', set the redis uri
    #: redis://username:password@host:port/db
    CACHE_REDIS_URL = ""

    #: CACHE_DIR
    #: Directory to store cache if CACHE_TYPE is filesystem, it will
    CACHE_DIR = ""

# ------------------------------------------------------------------------------
#: RECAPTCHA

    #: Flask-Recaptcha
    #: Register your application at https://www.google.com/recaptcha/admin

    #: RECAPTCHA_ENABLED
    RECAPTCHA_ENABLED = False

    #: RECAPTCHA_SITE_KEY
    RECAPTCHA_SITE_KEY = ""

    #: RECAPTCHA_SECRET_KEY
    RECAPTCHA_SECRET_KEY = ""


# ------------------------------------------------------------------------------
#: INSTALLED_APPS
    #
    # Installed apps/extensions to your application
    # The order is important. If an app depends on another app to work
    # that must be placed before the calling app
    #
    INSTALLED_APPS = [
        AVAILABLE_APPS["ERROR_PAGE"]
    ]

# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------
# --- ENVIRONMENT BASED CONFIG -------------------------------------------------


class Dev(BaseConfig):
    """
    Config for development environment
    """
    SERVER_NAME = None
    DEBUG = True
    SECRET_KEY = "PLEASE CHANGE ME"


class Prod(BaseConfig):
    """
    Config for Production environment
    """
    SERVER_NAME = None
    DEBUG = False
    SECRET_KEY = None
    COMPRESS_HTML = True



"""
________________________________________________________________________________

 /$$   /$$                                            /$$
| $$  | $$                                           | $$
| $$  | $$  /$$$$$$   /$$$$$$  /$$$$$$  /$$$$$$/$$$$ | $$$$$$$   /$$$$$$
| $$$$$$$$ |____  $$ /$$__  $$|____  $$| $$_  $$_  $$| $$__  $$ /$$__  $$
| $$__  $$  /$$$$$$$| $$  \__/ /$$$$$$$| $$ \ $$ \ $$| $$  \ $$| $$$$$$$$
| $$  | $$ /$$__  $$| $$      /$$__  $$| $$ | $$ | $$| $$  | $$| $$_____/
| $$  | $$|  $$$$$$$| $$     |  $$$$$$$| $$ | $$ | $$| $$$$$$$/|  $$$$$$$
|__/  |__/ \_______/|__/      \_______/|__/ |__/ |__/|_______/  \_______/



https://github.com/mardix/harambe

________________________________________________________________________________
"""

import os
import re
import sys
import traceback
import logging
import importlib
import pkg_resources
import click
import yaml
import functools
import sh
import json
from werkzeug import import_string
from .__about__ import *
from harambe import utils
import flask
from .core import db

CWD = os.getcwd()
SKELETON_DIR = "skel"
APPLICATION_DIR = "%s/app" % CWD
APPLICATION_DATA_DIR = "%s/var" % APPLICATION_DIR
APPLICATION_BABEL_DIR = "%s/babel" % APPLICATION_DATA_DIR

class CLI(object):
    """
    For command line classes in which __init__ contains all the functions to use

    example

    class MyCLI(HarambeCLI):
        def __init__(self):

            @cli.command()
            def hello():
                click.echo("Hello world")

            @cli.command()
            @click.argument("name")
            def say_my_name(name):
                click.echo("My name is %s" % name)
    """
    pass


def get_project_dir_path(project_name):
    return "%s/%s" % (APPLICATION_DIR, project_name)


def copy_resource_dir(src, dest):
    """
    To copy package data directory to destination
    """
    package_name = "harambe"
    dest = (dest + "/" + os.path.basename(src)).rstrip("/")
    if pkg_resources.resource_isdir(package_name, src):
        if not os.path.isdir(dest):
            os.makedirs(dest)
        for res in pkg_resources.resource_listdir(__name__, src):
            copy_resource_dir(src + "/" + res, dest)
    else:
        if not os.path.isfile(dest) and os.path.splitext(src)[1] not in [".pyc"]:
            copy_resource_file(src, dest)


def copy_resource_file(src, dest):
    with open(dest, "wb") as f:
        f.write(pkg_resources.resource_string(__name__, src))

# ------------------------------------------------------------------------------
application = None


def get_propel_config(cwd, key):
    with open("%s/%s" % (cwd, "propel.yml")) as f:
        config = yaml.load(f)
    return config.get(key)


def get_push_remotes_list(cwd, key=None):
    """
    Returns the remote hosts in propel
    :param cwd:
    :param key:
    :param file:
    :return: list
    """
    config = get_propel_config(cwd, "deploy-remotes")
    return config[key] if key else [v for k, l in config.items() for v in l]


def git_push_to_master(cwd, hosts, name="all", force=False):
    """
    To push to master
    :param cwd:
    :param hosts:
    :param force:
    :return:
    """

    def process_output(line):
        print(line)

    with sh.pushd(cwd):
        name = "harambe_deploy_%s" % name

        if sh.git("status", "--porcelain").strip():
            raise Exception("Repository is UNCLEAN. Commit your changes")

        remote_list = sh.git("remote").strip().split()
        if name in remote_list:
            sh.git("remote", "remove", name)
        sh.git("remote", "add", name, hosts[0])

        if len(hosts) > 1:
            for h in hosts:
                sh.git("remote", "set-url", name, "--push", "--add", h)

        _o = ["push", name, "master"]
        if force:
            _o.append("--force")
        sh.git(*_o, _out=process_output)
        sh.git("remote", "remove", name)


# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------
def catch_exception(func):
    @functools.wraps(func)
    def decorated_view(*args, **kwargs):
        try:
            func(*args, **kwargs)
        except Exception as ex:
            print("-" * 80)
            print("Exception Uncaught")
            print("-" * 80)
            print(ex)
            print("-" * 80)
    return decorated_view


def cwd_to_sys_path():
    sys.path.append(CWD)


def project_name(name):
    return re.compile('[^a-zA-Z0-9_]').sub("", name)


def header(title=None):
    print(__doc__)
    print("v. %s" % __version__)
    print("_" * 80)
    if title:
        print("** %s **" % title)


def build_assets(app):
    from webassets.script import CommandLineEnvironment
    assets_env = app.jinja_env.assets_environment
    log = logging.getLogger('webassets')
    log.addHandler(logging.StreamHandler())
    log.setLevel(logging.DEBUG)
    cmdenv = CommandLineEnvironment(assets_env, log)
    cmdenv.build()


@click.group()
def cli():
    """ Harambe (http://mardix.github.io/harambe) """

@click.group()
def cli_user():
    """ Harambe (http://mardix.github.io/harambe) """


@cli.command("init")
@catch_exception
def init():
    """  Setup Harambe in the current directory """

    harambepyfile = os.path.join(os.path.join(CWD, "harambe_init.py"))

    header("Initializing Harambe ...")
    if os.path.isfile(harambepyfile):
        print("WARNING: It seems like Harambe is already setup!")
        print("*" * 80)
    else:
        print("")
        print("Copying files to the current directory...")
        copy_resource_dir(SKELETON_DIR + "/create/", CWD)
        print("")

        _npm_install_static()
        print("")
        print("----- We did it for Harambe! ----")
        print("")
        print("- Harambe is now setup")
        print("")
        print("> What's next?")
        print("- Edit the config [ app/config.py ] ")
        print("- If necessary edit and run the command [ harambe db-sync ]")
        print("- Launch app on development mode, run [ harambe serve ]")
        print("")
        print("*" * 80)

@cli.command("addview")
@click.argument("name")
@click.option("--no-template", "-t", is_flag=True, default=False)
def add_view(name, no_template):
    """ Create a new view and template page """

    app_dest = os.path.join(CWD, "app")
    viewsrc = "%s/create-view/view.py" % SKELETON_DIR
    tplsrc = "%s/create-view/template.jade" % SKELETON_DIR
    viewdest_dir = os.path.join(app_dest, "views")
    viewdest = os.path.join(viewdest_dir, "%s.py" % name)
    tpldest_dir = os.path.join(app_dest, "templates/%s/Index" % name)
    tpldest = os.path.join(tpldest_dir, "index.jade")

    header("Adding New View")
    print("View: %s" % viewdest.replace(CWD, ""))
    if not no_template:
        print("Template: %s" % tpldest.replace(CWD, ""))
    else:
        print("* Template will not be created because of the flag --no-template| -t")
    if os.path.isfile(viewdest) or os.path.isfile(tpldest):
        print("*** ERROR: View or Template file exist already")
    else:
        if not os.path.isdir(viewdest_dir):
            utils.make_dirs(viewdest_dir)
        copy_resource_file(viewsrc, viewdest)

        if not no_template:
            if not os.path.isdir(tpldest_dir):
                utils.make_dirs(tpldest_dir)
            copy_resource_file(tplsrc, tpldest)

    print("")
    print("*" * 80)

@cli.command("serve")
@click.option("--port", "-p", default=5000)
@catch_exception
def server(port):
    """ Serve application in development mode """

    header("Serving application in development mode ... ")
    print("")
    print("- Port: %s" % port)
    print("")
    cwd_to_sys_path()
    application.app.run(debug=True, host='0.0.0.0', port=port)

@cli.command("db-sync")
def db_sync():
    """ Sync database Create new tables etc... """

    print("Syncing up database...")
    cwd_to_sys_path()
    if application.app.db and application.app.db.Model:
        application.app.db.create_all()
        for m in application.app.db.Model.__subclasses__():
            if hasattr(m, "_syncdb"):
                print("Sync up model: %s ..." % m.__name__)
                getattr(m, "_syncdb")()

    print("Done")

def _set_flask_alembic():
    from flask_alembic import Alembic

    """ Add the SQLAlchemy object in the global extension """
    application.app.extensions["sqlalchemy"] = type('', (), {"db": db})
    alembic = Alembic()
    alembic.init_app(application.app)
    return alembic



#@cli.command("db-migrate")
def db_migrate():
    """Migrate DB """
    print("Not ready for use")
    exit()
    cwd_to_sys_path()
    alembic = _set_flask_alembic()
    with application.app.app_context():
        p = application.app.db.Model.__subclasses__()
        print(p)

        # Auto-generate a migration
        alembic.revision('making changes')

        # Upgrade the database
        alembic.upgrade()


@cli.command("assets2s3")
@catch_exception
def assets2s3():
    """ Upload assets files to S3 """
    import flask_s3

    header("Assets2S3...")
    print("")
    print("Building assets files..." )
    print("")
    build_assets(application.app)
    print("")
    print("Uploading assets files to S3 ...")
    flask_s3.create_all(application.app)
    print("")


@cli.command("deploy")
@click.option("--site", "-s", default=None)
@click.option("--remote", "-r", default=None)
@click.option("--force", "-f", is_flag=True, default=False)
@catch_exception
def deploy(site, remote, force):
    """ To deploy application to a git repository """

    if site:
        pass
    else:
        remote_name = remote or "ALL"
        print("Deploying application to remote: %s " % remote_name)
        print("...")
        hosts = get_push_remotes_list(CWD, remote or None)
        git_push_to_master(cwd=CWD, hosts=hosts, name=remote_name, force=force)
        print("Done!")


@cli.command("version")
def version():
    print("-" * 80)
    print(__version__)
    print("-" * 80)

# @cli.command("babel-build")
# def babelbuild():
#     print("Babel Build....")
#     #sh.pybabel()

def _npm_install_static():
    print("NPM Install packages...")
    static_path = os.path.join(CWD, "app/static")
    package_json = os.path.join(static_path, "package.json")

    try:
        if os.path.isfile(package_json):
            with sh.pushd(static_path):
                sh.npm("install", "-f")
        else:
            print("**ERROR: Can't install static files, `package.json` is missing at `%s`" % package_json)
            print("*" * 80)
    except sh.CommandNotFound as e:
        print("")
        print("*** Error Command Not Found: `{0}` is not found. You need to install `{0}` to continue".format(str(e)))
        print("*" * 80)

@cli.command("npm-install")
def install_static():
    """ Install NPM Packages for the front end in the /static/ dir"""
    _npm_install_static()


# ------------------------------------------------------------------------------
# ------------------------------------------------------------------------------

def cmd():
    """
    Help to run the command line
    :return:
    """
    global application

    harambepyfile = os.path.join(os.path.join(CWD, "harambepy.py"))
    if os.path.isfile(harambepyfile):
        cwd_to_sys_path()
        application = import_string("harambepy")
    else:
        print("-" * 80)
        print("** Missing << 'harambepy.py' >> @ %s" % CWD)
        print("-" * 80)

    print(sys.argv)
    if "harambe" in sys.argv[0].split("/")[-1]:
        [cmd(cli_user.command, click) for cmd in CLI.__subclasses__()]
        cli_user()
    else:
        cli()

